{-----------------------------------------------------------------------------
 Unit Name: CocoProductions
 Author:    Administrator
Purpose: Container for GUI-model of ATG File - a list of productions
-----------------------------------------------------------------------------}


unit CocoProductions;

interface

uses
  Windows, Messages, SysUtils, Classes, CocoProduction,
  CocoProductionPath, pbStrLib, jclSysUtils;

type
  TCocoProductions = class(TComponent)
  private
    FCompilerName: string;
    FIsEditing: boolean;
    FIsEditingPath: string;
    fsaMinIndent: Integer;
    procedure SetCompilerName(const Value: string);
    function GetCount: Integer;
    procedure SetIsEditing(const Value: boolean);
    procedure SetIsEditingPath(const Value: string);
    { Private declarations }
  protected
    { Protected declarations }
    FProdList: TList;
    procedure SortProductions;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddProduction(const ProdName, ProdBody: string): TCocoProduction;
    procedure RemoveProduction(const ProdName: string);
    function GetProduction(Idx: Integer): TCocoProduction;
    function GetProductionTitled(const sProdTitle: string): TCocoProduction;
    procedure GetProductionNames(SL: TStrings);
    function GetATG: string;
  published
    { Published declarations }
    property CompilerName: string read FCompilerName write SetCompilerName;
    property Count: Integer read GetCount;
    property IsEditing: boolean read FIsEditing write SetIsEditing;
    property IsEditingPath: string read FIsEditingPath write SetIsEditingPath;
    property saMinIndent: Integer read fsaMinIndent write fsaMinIndent;
  end;

implementation

{ TCocoProductions }

function TCocoProductions.AddProduction(const ProdName,
  ProdBody: string): TCocoProduction;
var
  p: TCocoProduction;
begin
  p := TCocoProduction.Create(Self);
  P.ProductionName := ProdName;
  //  P.AddPath(ProdBody);
  P.ProdIndex := FProdList.Count;
  FProdList.Add(p);
  Result := P;
end;

constructor TCocoProductions.Create(AOwner: TComponent);
begin
  inherited;
  FProdList := TList.Create;
end;

destructor TCocoProductions.Destroy;
begin
  FreeObjectList(FProdList);
  inherited;
end;

function TCocoProductions.GetATG: string;
var
  idx: integer;
  p: TCocoProduction;
begin
  if not FIsEditing then
  begin
    Result := 'PRODUCTIONS' + CRLF;
    //  SortProductions;
    for idx := 0 to FProdList.Count - 1 do
    begin
      P := FProdList[idx];
      P.saMinIndent := fsaMinIndent;
      if not (p.ProductionName = FIsEditingPath) then
        Result := Result + P.GetATG
      else
        Result := Result + ModTextTag + CRLF;
    end;
  end
  else
  begin
    // Editing entire set of productions
    Result := 'PRODUCTIONS' + CRLF + ModTextTag + CRLF;
  end;
end;


function TCocoProductions.GetCount: Integer;
begin
  Result := FProdLIst.Count;
end;

function TCocoProductions.GetProduction(Idx: Integer): TCocoProduction;
begin
  Result := FProdList[idx];
end;

procedure TCocoProductions.GetProductionNames(SL: TStrings);
var
  idx: integer;
  P: TCocoProduction;
begin
  SL.Clear;
  for idx := 0 to FProdList.Count - 1 do
  begin
    P := FProdLIst[idx];
    SL.Add(P.ProductionName);
  end;
end;

function TCocoProductions.GetProductionTitled(const sProdTitle: string): TCocoProduction;
var
  idx: integer;
  S: string;
  P: TCocoProduction;
begin
  Result := nil;
  for idx := 0 to FProdList.Count - 1 do
  begin
    P := FProdList[idx];
    P.saMinIndent := fsaMinIndent;
    S := P.ProductionName;
    if (AnsiSameText(S, sProdTitle)) then
    begin
      Result := P;
      Exit;
    end;
  end;
end;


procedure TCocoProductions.RemoveProduction(const ProdName: string);
var
  idx: integer;
  p: TCocoProduction;
begin
  for idx := 0 to FProdList.Count - 1 do
  begin
    P := FProdList[idx];
    if (SameText(P.ProductionName, ProdName)) then
    begin
      FProdList.Delete(idx);
      if Assigned(P) then
        P.Free;
    end;
  end;
end;

procedure TCocoProductions.SetCompilerName(const Value: string);
begin
  FCompilerName := Value;
end;

procedure TCocoProductions.SetIsEditing(const Value: boolean);
begin
  FIsEditing := Value;
end;

procedure TCocoProductions.SetIsEditingPath(const Value: string);
begin
  FIsEditingPath := Value;
end;

{-----------------------------------------------------------------------------
  Procedure: TCocoProductions.SortProductions

  SortProductions orders the production objects such that
  the ATG file can be simply spat out in the correct manner.

  The rules of the sort are these:
  - The goal production is always first.
  - Productions that mention no other productions are flagged as going last
  - Other Productions are placed unsorted below the goal production
  -     Then each one is studied to see if it is mentioned in the production
        below. If YES, then it is moved down 1 position and reassessed until
        it is either not mentioned in the production below or it is at the end of the
        list.
        Then move on to the next item below the goal symbol that is yet to be sorted.
        If the previous analysis moved an item, then start at the top.
  - Store the name of each production that gets sorted, such that it does not get sorted twice.
-----------------------------------------------------------------------------}

procedure TCocoProductions.SortProductions;
var
  sGoalProdName: string;
  slSortedProds: TStringList;
  bProdWasMoved: boolean;
  CurrentStartPosition, SubjectProdIdx, CmpProdCnt: integer;
  PSubject, PCmp: TCocoProduction;
begin
  if (FProdList.Count < 2) then
    // we don't need to do any sorting
    Exit;

  { TODO : Run a test on blank compiler names etc to verify good error handling }
  Assert(FCompilerName <> '', 'Bug: Pat, Catch this one upstream!');

  sGoalProdName := FCompilerName;
  slSortedProds := TStringList.Create;
  bProdWasMoved := False;
  // Find Goal Production
  PSubject := GetProductionTitled(sGoalProdName);
  if (PSubject = nil) then
  begin
    raise Exception.Create('Error building ATG file: One production name must match the compiler name.');
  end
  else
  begin
    SubjectProdIdx := FProdList.IndexOf(PSubject);
    if (SubjectProdIdx > 0) then
      FProdList.Exchange(SubjectProdIdx, 0);
  end;

  // Start Sorting the rest
  CurrentStartPosition := 1;
  SubjectProdIdx := 1;
  CmpProdCnt := 1;

  while (SubjectProdIdx < FProdList.Count - 1) do
  begin
    // Scan the atg of PCmp to see if it contains title of PSubject
    PSubject := FProdList[SubjectProdIdx];
    PCmp := FProdList[SubjectProdIdx + CmpProdCnt];

    if (slSortedProds.IndexOf(Psubject.ProductionName) > 1) then
    begin
      if (not bProdWasMoved) then
        inc(CurrentStartPosition);
      SubjectProdIdx := CurrentStartPosition;
      bProdWasMoved := False;
      CmpProdCnt := 1;
    end
    else if (pos(UpperCase(Psubject.ProductionName), UpperCase(PCmp.GetATG)) > 0) then
    begin
      // demote PSubject to below PCmp
      FProdList.Exchange(SubjectProdIdx, SubjectProdIdx + CmpProdCnt);
      Inc(SubjectProdIdx); // as we are going to follow this production in another cycle
      bProdWasMoved := True;
    end
    else if (SubjectProdIdx = FProdList.Count - 2) then
    begin
      // Sorted
      // we leave this production where it is,
      // and add its name to the list of sorted productions
      slSortedProds.Add(Psubject.ProductionName);
      if (not bProdWasMoved) then
        inc(CurrentStartPosition);
      SubjectProdIdx := CurrentStartPosition;
      bProdWasMoved := False;
      CmpProdCnt := 1;
    end
    else
    begin
      if (SubjectProdIdx + CmpProdCnt < FProdList.Count - 1) then
        inc(CmpProdCnt)
      else
      begin
        slSortedProds.Add(Psubject.ProductionName);
        if (not bProdWasMoved) then
          inc(CurrentStartPosition);
        SubjectProdIdx := CurrentStartPosition;
        bProdWasMoved := False;
        CmpProdCnt := 1;
      end;
    end;
  end;

  slSortedProds.Free;
end;

end.
